"""
Forms definition
"""

# Django
from django import forms
from django.contrib.auth.models import Group
from django.forms import ModelForm
from django.utils.translation import gettext_lazy as _

# ckEditor
from ckeditor_uploader.widgets import CKEditorUploadingWidget

# AA Bulletin Board
from aa_bulletin_board.models import Bulletin


class SpecialModelChoiceIterator(forms.models.ModelChoiceIterator):
    """
    Variant of Django's ModelChoiceIterator to prevent it from always re-fetching the
    given queryset from database.
    """

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)

        queryset = self.queryset

        for obj in queryset:
            yield self.choice(obj)


class SpecialModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    """
    Variant of Django's ModelMultipleChoiceField to prevent it from always
    re-fetching the given queryset from database.
    """

    iterator = SpecialModelChoiceIterator

    def _get_queryset(self):
        return self._queryset

    def _set_queryset(self, queryset):
        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)


class BulletinForm(ModelForm):
    """
    Form for bulletins
    """

    title = forms.CharField()

    content = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_bulletin_board",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        )
    )

    groups = SpecialModelMultipleChoiceField(
        required=False,
        queryset=Group.objects.all(),
        help_text=_(
            "Restrict this bulletin to certain groups. "
            "If no group restrictions are in place, everyone who has access to "
            "this module can read this bulletin."
        ),
    )

    def __init__(self, *args, **kwargs):
        groups_queryset = kwargs.pop("groups_queryset", None)

        super().__init__(*args, **kwargs)

        if groups_queryset:
            self.fields["groups"].queryset = groups_queryset

    class Meta:
        model = Bulletin
        fields = ["title", "content", "groups"]



# Generated by Django 3.1.7 on 2021-04-04 11:16

# Django
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

# ckEditor
import ckeditor_uploader.fields

# AA Bulletin Board
import aa_bulletin_board.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "Bulletins",
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("manage_bulletins", "Can manage (add/change/remove) bulletins"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Bulletin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("slug", models.CharField(max_length=255)),
                (
                    "content",
                    ckeditor_uploader.fields.RichTextUploadingField(
                        blank=True, null=True
                    ),
                ),
                (
                    "created_date",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("updated_date", models.DateTimeField(blank=True, null=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=models.SET(
                            aa_bulletin_board.models.get_sentinel_user
                        ),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Bulletin",
                "verbose_name_plural": "Bulletins",
                "default_permissions": (),
            },
        ),
    ]



"""
App config
"""

# Django
from django.apps import AppConfig

# AA Bulletin Board
from aa_bulletin_board import __version__


class AaBulletinBoardConfig(AppConfig):
    """
    Application config
    """

    name = "aa_bulletin_board"
    label = "aa_bulletin_board"
    verbose_name = f"Bulletin Board v{__version__}"



"""
Test our helper functions
"""

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.test import TestCase

from ..models import Bulletin
from .utils import create_fake_user

fake = Faker()


class TestHelpers(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(1001, "Peter Parker")

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            1002, "Bruce Wayne", permissions=["aa_bulletin_board.basic_access"]
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            1003,
            "Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_translate_russian_letters_in_slug(self):
        """
        Test that russian letters in a slug are translated
        :return:
        :rtype:
        """

        bulletin = Bulletin.objects.create(
            title="дрифтерке, рорки в док",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.assertTrue(bulletin.slug, "drifterke-rorki-v-dok")



"""
Test for the UI
"""

# Third Party
from django_webtest import WebTest
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.urls import reverse

from ..models import Bulletin
from .utils import create_fake_user

fake = Faker()


class TestBulletinUI(WebTest):
    @classmethod
    def setUpClass(cls):
        """
        Set up groups and users
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(1001, "Peter Parker")

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            1002, "Bruce Wayne", permissions=["aa_bulletin_board.basic_access"]
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            1003,
            "Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_show_bulletin_page(self):
        """
        Test if a bulletin is shown
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Test Bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(self.user_1002)

        # when
        page = self.app.get(
            reverse("aa_bulletin_board:view_bulletin", args=[bulletin.slug])
        )

        # then
        self.assertTemplateUsed(page, "aa_bulletin_board/bulletin.html")

    def test_should_redirect_to_bulletin_dashboard_when_bulletin_does_not_exist(self):
        """
        Test should redirect to bulletin dashboard when bulletin does not exist
        :return:
        :rtype:
        """

        # given
        self.app.set_user(self.user_1002)

        # when
        page = self.app.get(reverse("aa_bulletin_board:view_bulletin", args=["foobar"]))

        # then
        self.assertRedirects(page, "/bulletin-board/")

    def test_should_show_create_bulletin_page(self):
        """
        Test if create bulletin is shown
        :return:
        :rtype:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_bulletin_board:create_bulletin"))

        # then
        self.assertTemplateUsed(page, "aa_bulletin_board/edit-bulletin.html")

    def test_should_show_edit_bulletin_page(self):
        """
        Test if edit bulletin is shown
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Test Bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(
            reverse("aa_bulletin_board:edit_bulletin", args=[bulletin.slug])
        )

        # then
        self.assertTemplateUsed(page, "aa_bulletin_board/edit-bulletin.html")

    def test_should_redirect_to_bulletin_dashboard_when_edit_bulletin_does_not_exist(
        self,
    ):
        """
        Test should redirect to bulletin dashboard when bulletin to edit does not exist
        :return:
        :rtype:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_bulletin_board:edit_bulletin", args=["foobar"]))

        # then
        self.assertRedirects(page, "/bulletin-board/")



"""
Testing access to bulletins
"""

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.test import TestCase
from django.urls import reverse

from ..models import Bulletin
from .utils import create_fake_user

fake = Faker()


class TestBulletin(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(1001, "Peter Parker")

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            1002, "Bruce Wayne", permissions=["aa_bulletin_board.basic_access"]
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            1003,
            "Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_show_dashboard(self):
        """
        Test that a user with basic_access can see the bulletin board
        :return:
        :rtype:
        """

        # given
        self.client.force_login(self.user_1002)

        # when
        res = self.client.get(reverse("aa_bulletin_board:dashboard"))

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_bulletin(self):
        """
        Test that a user with basic_access can see the bulletin
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(self.user_1002)

        # when
        res = self.client.get(
            reverse("aa_bulletin_board:view_bulletin", kwargs={"slug": bulletin.slug})
        )
        result = Bulletin.objects.user_has_access(self.user_1002)

        # then
        self.assertEqual(res.status_code, 200)
        self.assertIn(bulletin, result)

    def test_should_not_show_dashboard(self):
        """
        Test that a user without basic_access can't see the bulletin board
        :return:
        :rtype:
        """

        # given
        self.client.force_login(self.user_1001)

        # when
        res = self.client.get(reverse("aa_bulletin_board:dashboard"))

        # then
        self.assertIsNot(res.status_code, 200)

    def test_should_not_show_bulletin(self):
        """
        Test that a user without basic_access can't see the bulletin
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(self.user_1001)

        # when
        res = self.client.get(
            reverse("aa_bulletin_board:view_bulletin", kwargs={"slug": bulletin.slug})
        )

        # then
        self.assertIsNot(res.status_code, 200)

    def test_should_return_bulletin_for_user_with_perm_manage_bulletins(self):
        """
        Test that a user with "aa_bulletin_board.manage_bulletins" can see all bulletins
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        # when
        result = Bulletin.objects.user_has_access(self.user_1003)

        # then
        self.assertIn(bulletin, result)

    def test_should_return_restricted_bulletin_for_user_with_perm_manage_bulletins(
        self,
    ):
        """
        Test that a user with "aa_bulletin_board.manage_bulletins"
        can see all bulletins, even restricted
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        bulletin.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(self.user_1003)

        # then
        self.assertIn(bulletin, result)

    def test_should_return_bulletin_with_no_groups(self):
        """
        Test that any user with access can se non restricted bulletins
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        # when
        result = Bulletin.objects.user_has_access(self.user_1002)

        # then
        self.assertIn(bulletin, result)

    def test_should_return_bulletin_for_group_member(self):
        """
        Test that group restricted bulletins are visible for users who have this group
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        bulletin.groups.add(self.group)
        self.user_1001.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(self.user_1001)

        # then
        self.assertIn(bulletin, result)

    def test_should_not_return_bulletin_for_non_group_member(self):
        """
        Test that group restricted bulletins are not visible for
        users who don't have this group
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        bulletin.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(self.user_1002)

        # then
        self.assertNotIn(bulletin, result)



"""
Managers for our models
"""

# Django
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q


class BulletinQuerySet(models.QuerySet):
    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter boards that given user has access to.
        """

        # Forum manager have always access, so assign this permission wisely
        if user.has_perm("aa_bulletin_board.manage_bulletins"):
            return self

        # If not a forum manager, check if the user has access to the board
        return self.filter(
            Q(groups__in=user.groups.all()) | Q(groups__isnull=True)
        ).distinct()


class BulletinManagerBase(models.Manager):
    """
    BoardManagerBase
    """

    pass


BulletinManager = BulletinManagerBase.from_queryset(BulletinQuerySet)



"""
The models
"""

# Third Party
import unidecode

# Django
from django.contrib.auth.models import Group, User
from django.db import models, transaction
from django.utils.text import slugify
from django.utils.translation import gettext as _

# ckEditor
from ckeditor_uploader.fields import RichTextUploadingField

# AA Bulletin Board
from aa_bulletin_board.helpers import string_cleanup
from aa_bulletin_board.managers import BulletinManager


def get_sentinel_user() -> User:
    """
    Get user or create one
    :return:
    """

    return User.objects.get_or_create(username="deleted")[0]


def get_bulletin_slug_from_title(bulletin_title: str) -> str:
    """
    Get the slug from the title
    :param bulletin_title:
    :return:
    """

    run = 0
    bulletin_slug = slugify(unidecode.unidecode(bulletin_title), allow_unicode=True)

    while Bulletin.objects.filter(slug=bulletin_slug).exists():
        run += 1
        bulletin_slug = slugify(
            unidecode.unidecode(f"{bulletin_title}-{run}"), allow_unicode=True
        )

    return bulletin_slug


class General(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:
        verbose_name = "Bulletins"
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("manage_bulletins", "Can manage (add/change/remove) bulletins"),
        )


class Bulletin(models.Model):
    """
    Bulletin model
    """

    title = models.CharField(max_length=255)
    slug = models.SlugField(max_length=255, unique=True, allow_unicode=True)
    content = RichTextUploadingField(blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True, db_index=True)
    updated_date = models.DateTimeField(auto_now=True, null=True)
    created_by = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
    )
    groups = models.ManyToManyField(
        Group,
        blank=True,
        related_name="aa_bulletin_board_group_restriction",
    )

    objects = BulletinManager()

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Bulletin")
        verbose_name_plural = _("Bulletins")

    def __str__(self) -> str:
        return str(self.title)

    @transaction.atomic()
    def save(self, *args, **kwargs) -> None:
        """
        Add the slug on save
        """

        self.content = string_cleanup(self.content)

        if self.slug == "":
            bulletin_slug = get_bulletin_slug_from_title(bulletin_title=self.title)
            self.slug = bulletin_slug

        super().save(*args, **kwargs)



"""
Test the apps' template tags
"""

# Django
from django.template import Context, Template
from django.test import TestCase

# AA Bulletin Board
from aa_bulletin_board import __version__


class TestVersionedStatic(TestCase):
    def test_versioned_static(self):
        context = Context({"version": __version__})
        template_to_render = Template(
            "{% load aa_bulletin_board_versioned_static %}"
            "{% aa_bulletin_board_static 'aa_bulletin_board/css/aa-bulletin-board.min.css' %}"
        )

        rendered_template = template_to_render.render(context)

        self.assertInHTML(
            f'/static/aa_bulletin_board/css/aa-bulletin-board.min.css?v={context["version"]}',
            rendered_template,
        )



"""
Test bulletins
"""

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

from ..models import Bulletin
from .utils import create_fake_user

fake = Faker()


class TestBulletins(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(1001, "Peter Parker")

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            1002, "Bruce Wayne", permissions=["aa_bulletin_board.basic_access"]
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            1003,
            "Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_remove_bulletin(self):
        """
        Test if a bulletin should be removed
        :return:
        :rtype:
        """

        # given
        bulletin_title = fake.sentence()
        bulletin = Bulletin.objects.create(
            title=bulletin_title,
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.client.force_login(self.user_1003)

        # when
        response = self.client.get(
            reverse("aa_bulletin_board:remove_bulletin", args=[bulletin.slug])
        )
        messages = list(get_messages(response.wsgi_request))

        # then
        self.assertRedirects(response, "/bulletin-board/")
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), f'Bulletin "{bulletin_title}" deleted.')

    def test_should_raise_does_not_exist_exception_when_delete_bulletin_not_found(self):
        """
        Test if a bulletin that doesn't exist should be removed
        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title=fake.sentence(),
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(self.user_1003)

        # when
        response = self.client.get(
            reverse("aa_bulletin_board:remove_bulletin", args=["foobarsson"])
        )
        messages = list(get_messages(response.wsgi_request))

        self.assertRaises(bulletin.DoesNotExist)
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            str(messages[0]),
            "The bulletin you are trying to delete for does not exist.",
        )
